﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {
  [StorableClass]
  [Item("NormalizedHypervolumeAnalyzer", "Computes the enclosed Hypervolume between the current front and a given reference Point")]
  public class NormalizedHypervolumeAnalyzer : MOTFAnalyzer {
    private const string bestKnownFront = "BestKnownFront Zitzler";
    private const string resultsHV = "NormalizedHypervolume";
    private const string resultsDist = "Absolute Distance to Normalized BestKnownHypervolume";

    private const string bestknownHV = "NormalizedBestKnownHyperVolume";

    public IValueParameter<DoubleMatrix> OptimalFrontParameter {
      get { return (IValueParameter<DoubleMatrix>)Parameters[bestKnownFront]; }
    }

    public IValueParameter<DoubleValue> BestKnownHyperVolumeParameter {
      get { return (IValueParameter<DoubleValue>)Parameters[bestknownHV]; }
    }


    public NormalizedHypervolumeAnalyzer()
      : base() {
      Parameters.Add(new ValueParameter<DoubleMatrix>(bestKnownFront, "The true / best known pareto front"));
      Parameters.Add(new ValueParameter<DoubleValue>(bestknownHV, "The currently best known hypervolume"));
    }

    [StorableConstructor]
    protected NormalizedHypervolumeAnalyzer(bool deserializing) : base(deserializing) { }
    protected NormalizedHypervolumeAnalyzer(NormalizedHypervolumeAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new NormalizedHypervolumeAnalyzer(this, cloner);
    }


    private void RegisterEventHandlers() {
      OptimalFrontParameter.ValueChanged += OptimalFrontParameterOnValueChanged;
    }

    private void OptimalFrontParameterOnValueChanged(object sender, EventArgs e) {
      BestKnownHyperVolumeParameter.Value = new DoubleValue(0);
    }

    public override IOperation Apply() {
      var results = ResultsParameter.ActualValue;
      var qualities = QualitiesParameter.ActualValue;
      var testFunction = TestFunctionParameter.ActualValue;
      int objectives = qualities[0].Length;

      double best = BestKnownHyperVolumeParameter.Value.Value;
      if (OptimalFrontParameter.Value == null || OptimalFrontParameter.Value.Rows < 1 || OptimalFrontParameter.Value.Columns != qualities[0].Length) {
        return base.Apply(); // too pareto front nonexistant or with wrong number of dimensions
      }

      IEnumerable<double[]> front = NonDominatedSelect.SelectNonDominatedVectors(qualities.Select(q => q.ToArray()), testFunction.Maximization(objectives), true);

      if (!results.ContainsKey(resultsHV)) results.Add(new Result(resultsHV, typeof(DoubleValue)));
      if (!results.ContainsKey(resultsDist)) results.Add(new Result(resultsDist, typeof(DoubleValue)));
      if (!results.ContainsKey(bestknownHV)) results.Add(new Result(bestknownHV, typeof(DoubleValue)));
      else {
        DoubleValue dv = (DoubleValue)(results[bestknownHV].Value);
        best = dv.Value;
      }

      bool[] maximization = testFunction.Maximization(objectives);
      double[] invPoint = GetBestPoint(OptimalFrontParameter.Value, maximization);
      double[] refPoint = GetWorstPoint(OptimalFrontParameter.Value, maximization);
      double normalization = Hypervolume.Calculate(new double[][] { invPoint }, refPoint, maximization);
      double hv = front.Any() ? Hypervolume.Calculate(front, refPoint, maximization) / normalization : 0;

      if (double.IsNaN(best) || best < hv) {
        best = hv;
        BestKnownFrontParameter.ActualValue = new DoubleMatrix(MultiObjectiveTestFunctionProblem.To2D(qualities.Select(q => q.ToArray()).ToArray()));
      }

      results[resultsHV].Value = new DoubleValue(hv);
      results[bestknownHV].Value = new DoubleValue(best);
      results[resultsDist].Value = new DoubleValue(best - hv);

      return base.Apply();
    }


    private static double[] GetWorstPoint(DoubleMatrix value, bool[] maximization) {
      bool[] invMax = new bool[maximization.Length];
      int i = 0;
      foreach (bool b in maximization) {
        invMax[i++] = !b;
      }
      return GetBestPoint(value, invMax);
    }

    private static double[] GetBestPoint(DoubleMatrix value, bool[] maximization) {
      double[] res = new double[maximization.Length];
      for (int i = 0; i < maximization.Length; i++) {
        res[i] = maximization[i] ? Double.MinValue : Double.MaxValue;
      }

      for (int r = 0; r < value.Rows; r++) {
        for (int c = 0; c < maximization.Length; c++) {
          res[c] = maximization[c] ? Math.Max(res[c], value[r, c]) : Math.Min(res[c], value[r, c]);
        }
      }
      return res;
    }
  }
}
